import abc
import sys
import traceback
from typing import Any, Callable, Iterable, Type, TypeVar, Union


primitives = (bool, str, int, float, type(None))


def scrub_title_key(d: dict):
    """
    Helpful because pydantic schema dumps include an unecessary 'title' attribute;
    """
    d.pop("title", None)
    if d.get("type") == "object":
        assert "properties" in d
        for prop in d["properties"].keys():
            scrub_title_key(d["properties"][prop])
    return d


def is_primitive(obj):
    return isinstance(obj, primitives)


def get(obj: Any, *attr_path: Union[str, int]):
    cur = obj
    for key in attr_path:
        if cur is None:
            return None
        if type(cur) == list:
            if type(key) != int:
                return None
            if key >= len(cur):
                return None
            cur = cur[key]
        elif type(cur) == dict:
            if key not in cur:
                return None
            cur = cur[key]
        elif is_primitive(cur):
            # can't index a primitive
            return None
        else:
            if type(key) != str:
                return None
            cur = getattr(cur, key, None)
    return cur


T = TypeVar("T")
S = TypeVar("S")
E = TypeVar("E", bound=Exception)


class TracebackError(Exception, abc.ABC):
    def __init__(self, *args: object, tb: str = ""):
        pass


def try_or_raise(fn: Callable[[], T], exc: Type[TracebackError]) -> T:
    """
    Try 'fn()', raising 'exc' on any exception.
    """
    try:
        return fn()
    except Exception as e:
        exc_info = sys.exc_info()
        tb = "".join(traceback.format_exception(*exc_info))
        raise exc(e, tb=tb)
    return


def try_or(
    fn: Callable[[], T], on_err: Callable[[E], S], errtype: Type[E]
) -> Union[T, S]:
    """
    Try 'fn()', but call 'on_err' if an error of 'errtype' is thrown.
    """
    try:
        return fn()
    except errtype as e:
        return on_err(e)
    return


def map_multithreaded(fn, data: Iterable, max_threads: int) -> list:
    """
    Like builtin 'map', but executes across 'max_threads' threads, and returns a list not an iterator.
    Here, order will be preserved.
    """
    if not data:
        return []
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor(max_workers=max_threads) as executor:
        return list(executor.map(fn, data))
    return


def concurrently(*fns) -> list:
    """
    Call all provided functions in their own thread, returning the results in a list.
    """

    def call(fn):
        return fn()

    return map_multithreaded(call, fns, len(fns))



import logging
import os


def setup_logging():
    LOG_FMT = (
        f"%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s"
    )

    logging.basicConfig(
        level=getattr(logging, os.environ.get("LOG_LEVEL", "INFO")),
        format=LOG_FMT,
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )
    return



import abc
from collections import deque
import logging
import threading
import time
import urllib.parse

import requests


class RestClientBase:
    def __init__(
        self,
        base: str,
        *,
        headers: dict[str, str] | None = None,
        rate_limit_window_seconds: int | None = None,
        rate_limit_requests: int | None = None,
    ):
        self.base = base
        self.rate_limit_window_seconds = rate_limit_window_seconds
        self.rate_limit_requests = rate_limit_requests
        self._REQ_HISTORY_LOCK = threading.Lock()
        self._REQ_HISTORY: deque[float] = deque([])
        self.session = requests.Session()
        if headers:
            for k, v in headers.items():
                self.session.headers[k] = v
        return

    def _encode_query_params(self, params: dict) -> str:
        return "&".join(
            [f"{k}={urllib.parse.quote_plus(str(v))}" for k, v in params.items()]
        )

    def _wait_turn_for_request(self) -> None:
        if not self.rate_limit_requests or not self.rate_limit_window_seconds:
            return
        now = time.time()
        cutoff = now - self.rate_limit_window_seconds
        with self._REQ_HISTORY_LOCK:
            while self._REQ_HISTORY and self._REQ_HISTORY[0] < cutoff:
                self._REQ_HISTORY.popleft()
            if len(self._REQ_HISTORY) < self.rate_limit_requests:
                self._REQ_HISTORY.append(now)
                return
        # otherwise, too many in-flight requests
        time.sleep(self.rate_limit_window_seconds)
        self._wait_turn_for_request()
        return

    def _req(self, method: str, path: str, *, ignore_error: bool = False, **kwargs):
        url = f"{self.base}{path}"
        self._wait_turn_for_request()
        res = self.session.request(method, url, **kwargs)
        if not res.ok:
            logging.error("request to %s failed: %s", url, res.text)
        if not ignore_error:
            res.raise_for_status()
        return res



import argparse
from typing import (
    Callable,
    NamedTuple,
    Type,
    TypeVar,
)

BaseArgs = NamedTuple


T = TypeVar("T", bound=BaseArgs)


def make_cli_arg_parser(
    *,
    name: str,
    desc: str,
    args_shape: Type[T],
) -> Callable[[], T]:
    type_map = args_shape.__annotations__
    defaults = args_shape._field_defaults

    parser = argparse.ArgumentParser(
        prog=name,
        description=desc,
    )
    for name, dtype in type_map.items():
        cli_argname = name.replace("_", "-")
        kwargs = {
            "dest": name,
            "required": name not in defaults,
            "type": dtype,
        }
        if name in defaults:
            kwargs["default"] = defaults[name]
        parser.add_argument(f"--{cli_argname}", **kwargs)

    def parse_args() -> T:
        args = parser.parse_args()
        return args_shape(**{k: v for k, v in vars(args).items() if v})  # type: ignore

    return parse_args



import base64
import json
import os
import subprocess
from typing import (
    Optional,
)

import boto3  # type: ignore
from botocore.config import Config  # type: ignore


DEFAULT_REGION = "us-east-2"


class AwsServiceSession:
    def __init__(
        self,
        service_name: str,
        *,
        profile: str | None = None,
        region_name: str | None = None,
        creds_bundle: Optional[str] = None,
    ):
        if not profile:
            profile = os.environ.get("AWS_PROFILE")
        if not region_name:
            region_name = os.environ.get("AWS_REGION", DEFAULT_REGION)
        self.service_name = service_name
        if creds_bundle:
            creds = json.loads(base64.b64decode(creds_bundle))
            self.aws_session = self._init_aws_session(**creds)
        else:
            self.aws_session = self._init_aws_session(profile_name=profile)
        self.service = self._init_connection(region_name)
        return

    def _init_aws_session(
        self,
        **kwargs,
    ) -> boto3.session.Session:
        return boto3.session.Session(**kwargs)

    def _init_connection(self, region: Optional[str]):
        if region is not None:
            return self.aws_session.client(
                self.service_name, config=Config(region_name=region)
            )
        return self.aws_session.client(self.service_name)



#!/usr/bin/env python

from setuptools import find_packages, setup



setup(
    name="btdcore",
    version="v0.1.1",
    description="Core Python library for Black Tusk Data",
    author="Liam Tengelis",
    author_email="liam.tengelis@blacktuskdata.com",
    packages=find_packages(),
    package_data={
        "": ["*.yaml"],
        "btdcore": [
            "py.typed",
        ],
    },
    install_requires=[
        "pydantic",
        "requests",
    ],
)



import dataclasses as dc
import logging
import traceback
from typing import (
    Any,
    Optional,
    List,
)

import boto3  # type: ignore
from botocore.config import Config  # type: ignore

from .core import AwsServiceSession

AWSDbCon = Any


def parse_data(data: dict):
    try:
        key, val = next(iter(data.items()))
        if key == "S":
            return val
        if key == "N":
            return float(val)
        if key == "L":
            return [parse_data(r) for r in val]
        if key == "M":
            return parse_row(val)
        if key == "BOOL":
            return val
        # I guess this is like a 'null' marker?
        if key == "NULL":
            return val
        else:
            logging.error("UNKNOWN KEY: %s WITH VAL: %s IN OBJ %s", key, val, data)
    except Exception as e:
        logging.error("caught error: %s", e)
    return


def parse_row(row: dict):
    return {k: parse_data(data) for k, data in row.items()}


@dc.dataclass
class TableCon:
    dbcon: AWSDbCon
    table_name: str

    def get_item(self, key_exp: dict):
        return self.dbcon.get_item(
            TableName=self.table_name,
            Key=key_exp,
        )["Item"]

    def get_items(self, key_exprs: List[dict]):
        res = self.dbcon.batch_get_item(
            RequestItems={
                self.table_name: {
                    "Keys": key_exprs,
                }
            }
        )
        if res.get("UnprocessedKeys"):
            # TODO: handle this by simply re-fetching the unprocessed keys
            raise Exception("failed to fetch all items from DynamoDB")
        return res["Responses"][self.table_name]

    def query(
        self,
        key_expr: str,
        values: dict,
        exp_attr_names: Optional[dict] = None,
        index_name: str | None = None,
    ):
        req = {
            "TableName": self.table_name,
            "KeyConditionExpression": key_expr,
            "ExpressionAttributeValues": values,
        }
        if index_name is not None:
            req["IndexName"] = index_name
        if exp_attr_names is not None:
            req["ExpressionAttributeNames"] = exp_attr_names
        res = self.dbcon.query(**req)
        items = res["Items"]
        last_evaluated_key = res.get("LastEvaluatedKey")
        while last_evaluated_key:
            logging.debug("another page...")
            res = self.dbcon.query(
                **req,
                **{
                    "ExclusiveStartKey": last_evaluated_key,
                },
            )
            last_evaluated_key = res.get("LastEvaluatedKey")
            items.extend(res.get("Items", []))
        return items
        # return res["Items"]

    def put_item(self, item: dict):
        """Write an item to a table."""
        return self.dbcon.put_item(
            TableName=self.table_name,
            Item=item,
            ReturnValues="ALL_OLD",
        )

    def delete_item(self, key: dict):
        """Delete an item."""
        return self.dbcon.delete_item(
            TableName=self.table_name,
            Key=key,
        )

    def describe_table(self):
        return self.dbcon.describe_table(TableName=self.table_name)

    def scan(self, page_size=40):
        desc = self.describe_table()
        item_count = desc["Table"]["ItemCount"]
        if not item_count:
            return []
        notify_progress = (page_size / item_count) < 0.1
        last_p_notify = 0
        if notify_progress:
            logging.info("fetching %d records", item_count)
        rows = []
        res = self.dbcon.scan(
            TableName=self.table_name,
            Limit=page_size,
        )
        rows.extend(res["Items"])
        try:
            while res.get("LastEvaluatedKey"):
                res = self.dbcon.scan(
                    TableName=self.table_name,
                    ExclusiveStartKey=res["LastEvaluatedKey"],
                    Limit=page_size,
                )
                rows.extend(res["Items"])
                p_done = len(rows) * 100 / item_count
                p_done_to_tenth = int(10 * p_done) // 10
                if notify_progress:
                    if p_done_to_tenth > last_p_notify:
                        logging.info(f"{p_done_to_tenth}% done")
                        last_p_notify = p_done_to_tenth
        except KeyboardInterrupt:
            logging.info("exiting early...")
            return rows
        return rows


class AWSDynamoCon:
    """
    An interface into dynamodb.
    """

    def __init__(self):
        """Set up AWS and dynamo connection."""
        self.dbcon = AwsServiceSession("dynamodb").service
        self.all_tables = []
        return

    def get_table_name(self, name: str, suffix: str):
        if not self.all_tables:
            self.all_tables = self.list_tables()
        matches = [t for t in self.all_tables if f"{name}-" in t and f"-{suffix}" in t]
        if matches:
            return matches[0]
        return

    def list_tables(self):
        tables = []
        kwargs = {}
        while True:
            res = self.dbcon.list_tables(**kwargs)
            tables.extend(res["TableNames"])
            if "LastEvaluatedTableName" in res:
                kwargs["ExclusiveStartTableName"] = res["LastEvaluatedTableName"]
            else:
                return tables
        return

    def connect_table(self, table_name: str):
        return TableCon(self.dbcon, table_name)
